#define xdvi_mask_width 15
#define xdvi_mask_height 15
static short xdvi_mask_bits[] = {
   0x01c0, 0x03e0, 0x07f0, 0x0ff8,
   0x1ffc, 0x3ffe, 0x7fff, 0x7fff,
   0x7fff, 0x3ffe, 0x1ffc, 0x0ff8,
   0x07f0, 0x03e0, 0x01c0};
