#define xdvi_width 15
#define xdvi_height 15
#define xdvi_x_hot 7
#define xdvi_y_hot 7
static short xdvi_bits[] = {
   0x0080, 0x01c0, 0x03e0, 0x06b0,
   0x0c98, 0x188c, 0x3086, 0x7fff,
   0x3086, 0x188c, 0x0c98, 0x06b0,
   0x03e0, 0x01c0, 0x0080};
